pub struct OobScaleFpdec<I>(/* private fields */);Expand description
Out-of-band-scale fixed-point decimal.
I is the inner integer type. It could be signed i8, i16, i32,
i64, i128, or unsigned u8, u16, u32, u64, u128, with
about 2, 4, 9, 18/19 and 38 significant digits respectively.
For example, OobScaleFpdec<i64> means using i64 as the underlying
integer. It’s your job to save the out-of-band scale somewhere else.
The scale can be positive for fraction precision or be negative for omitting the low-order digits of integer values.
Compared to ConstScaleFpdec, this OobScaleFpdec has more verbose APIs:
- extra
diff_scaleargument for most operations such as*and/, but no need for+and-, - use
try_from_str()to convert from string with scale set, - use
(*, i32)tuple for converting from integers or floats, - use
to_f32()orto_f64()to convert to floats, - use
OobFmtforDisplayandFromStr, - no associate const
SCALE, - and others.
See the module-level documentation for more information.
Implementations
Sourceimpl<I> OobScaleFpdec<I>where
I: FpdecInner,
impl<I> OobScaleFpdec<I>where
I: FpdecInner,
Sourcepub const MAX_POWERS: Self
pub const MAX_POWERS: Self
The largest powers of 10.
Sourcepub fn checked_add(self, rhs: Self) -> Option<Self>
pub fn checked_add(self, rhs: Self) -> Option<Self>
Checked addition. Computes self + rhs, returning None if overflow occurred.
If you make sure no overflow error, you can use + and += instead
for convenience.
The right operand must have the same scale with self.
Sourcepub fn checked_sub(self, rhs: Self) -> Option<Self>
pub fn checked_sub(self, rhs: Self) -> Option<Self>
Checked subtraction. Computes self - rhs, returning None if overflow occurred.
If you make sure no overflow error, you can use - and -= instead
for convenience.
The right operand must have the same scale with self.
Sourcepub fn checked_mul_int(self, n: impl Into<I>) -> Option<Self>
pub fn checked_mul_int(self, n: impl Into<I>) -> Option<Self>
Checked multiplication with integer. Computes self * n, returning
None if overflow occurred.
If you make sure no overflow error, you can use * and *= instead
for convenience.
Sourcepub fn checked_mul_ratio<R>(self, a: R, b: R) -> Option<Self>where
R: IntoRatioInt<I>,
pub fn checked_mul_ratio<R>(self, a: R, b: R) -> Option<Self>where
R: IntoRatioInt<I>,
Computes self * a/b, returning None if overflow occurred.
Equivalent to Self::checked_mul_ratio_ext with Rounding::Round.
Sourcepub fn checked_mul_ratio_ext<R>(
self,
a: R,
b: R,
rounding: Rounding,
) -> Option<Self>where
R: IntoRatioInt<I>,
pub fn checked_mul_ratio_ext<R>(
self,
a: R,
b: R,
rounding: Rounding,
) -> Option<Self>where
R: IntoRatioInt<I>,
Computes self * a/b, returning None if overflow occurred.
The arguments a and b could be primitive integers,
ConstScaleFpdec or OobScaleFpdec with same scale.
Compared to calling Self::checked_mul_int and Self::checked_div_int
separately, this can avoid some potential overflow.
Examples:
// Here we use `ConstScaleFpdec` as example. It's same for `OobScaleFpdec`.
use primitive_fixed_point_decimal::{ConstScaleFpdec, fpdec};
type Balance = ConstScaleFpdec<i64, 2>;
type Quantity = ConstScaleFpdec<i32, 4>; // type for `a` and `b`
let margin: Balance = fpdec!(30000);
let deal: Quantity = fpdec!(0.2);
let total: Quantity = fpdec!(0.3);
assert_eq!(margin.checked_mul_ratio(deal, total).unwrap(), fpdec!(20000));
// integer
assert_eq!(margin.checked_mul_ratio(200, 300).unwrap(), fpdec!(20000));Sourcepub fn checked_div_int(self, n: impl Into<I>) -> Option<Self>
pub fn checked_div_int(self, n: impl Into<I>) -> Option<Self>
Checked division by integer, with Rounding::Round.
Computes self / n, returning None if n == 0 or overflow occurres.
If you make sure no overflow or 0-division error, you can use
/ and /= instead for convenience.
Sourcepub fn checked_div_int_ext(
self,
n: impl Into<I>,
rounding: Rounding,
) -> Option<Self>
pub fn checked_div_int_ext( self, n: impl Into<I>, rounding: Rounding, ) -> Option<Self>
Checked division by integer with rounding.
Computes self / n, returning None if n == 0 or overflow occurres.
Sourcepub const fn from_mantissa(i: I) -> Self
pub const fn from_mantissa(i: I) -> Self
Create a decimal from the underlying integer representation.
You must take care of the scale yourself.
This method is const, so it can be used to construct const values.
Examples:
// Here we use `ConstScaleFpdec` as example. It's same for `OobScaleFpdec`.
use primitive_fixed_point_decimal::{ConstScaleFpdec, fpdec};
type Dec = ConstScaleFpdec<i32, 4>; // scale is 4
assert_eq!(Dec::from_mantissa(123400), fpdec!(12.34));
// const values
const ONE: Dec = Dec::from_mantissa(1 * 10000);
const TEN: Dec = Dec::from_mantissa(10 * 10000);
const PI: Dec = Dec::from_mantissa(31400);
assert_eq!(ONE, fpdec!(1));
assert_eq!(TEN, fpdec!(10));
assert_eq!(PI, fpdec!(3.14));
// you can use `Dec::SCALE` for better consistency
const ONE_1: Dec = Dec::from_mantissa(10_i32.pow(Dec::SCALE as u32));
const TEN_1: Dec = Dec::from_mantissa(10_i32.pow(Dec::SCALE as u32 + 1));
const PI_1: Dec = Dec::from_mantissa(314 * 10_i32.pow(Dec::SCALE as u32 - 2));
assert_eq!(ONE_1, fpdec!(1));
assert_eq!(TEN_1, fpdec!(10));
assert_eq!(PI_1, fpdec!(3.14));Sourcepub const fn mantissa(self) -> I
pub const fn mantissa(self) -> I
Return the underlying integer representation.
You must take care of the scale yourself.
Examples:
// Here we use `ConstScaleFpdec` as example. It's same for `OobScaleFpdec`.
use primitive_fixed_point_decimal::{ConstScaleFpdec, fpdec};
type Dec = ConstScaleFpdec<i32, 4>; // scale is 4
let d: Dec = fpdec!(12.34);
assert_eq!(d.mantissa(), 123400);Sourcepub fn checked_mul<J>(
self,
rhs: OobScaleFpdec<J>,
diff_scale: i32,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
pub fn checked_mul<J>(
self,
rhs: OobScaleFpdec<J>,
diff_scale: i32,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
Checked multiplication.
Equivalent to Self::checked_mul_ext with Rounding::Round.
Sourcepub fn checked_mul_ext<J>(
self,
rhs: OobScaleFpdec<J>,
diff_scale: i32,
rounding: Rounding,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
pub fn checked_mul_ext<J>(
self,
rhs: OobScaleFpdec<J>,
diff_scale: i32,
rounding: Rounding,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
Checked multiplication. Computes self * rhs, returning None if
overflow occurred or argument diff_scale is out of range
[-Self::DIGITS, Self::DIGITS].
The type of rhs can have different inner integer J,
while the type of result must have the same I.
Argument: diff_scale = scale(self) + scale(rhs) - scale(result).
If the diff_scale < 0, then rounding operations are required and precision may be lost. You can specify the rounding type.
Examples
use primitive_fixed_point_decimal::{OobScaleFpdec, Rounding, fpdec};
type Balance = OobScaleFpdec<i64>;
type FeeRate = OobScaleFpdec<i16>; // different types
let balance: Balance = fpdec!(12.30, 2); // scale=2
let rate: FeeRate = fpdec!(0.01, 4); // scale=4
let fee: Balance = balance.checked_mul(rate, 4).unwrap();
assert_eq!(fee, fpdec!(0.12, 2));
let fee: Balance = balance.checked_mul_ext(rate, 4, Rounding::Ceiling).unwrap();
assert_eq!(fee, fpdec!(0.13, 2));Sourcepub fn checked_mul_const_scale<J, const S: i32>(
self,
rhs: ConstScaleFpdec<J, S>,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
pub fn checked_mul_const_scale<J, const S: i32>(
self,
rhs: ConstScaleFpdec<J, S>,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
Checked multiplication with ConstScaleFpdec.
Equivalent to Self::checked_mul_const_scale_ext with Rounding::Round.
If you make sure no overflow error, you can use * and *= instead for convenience.
Sourcepub fn checked_mul_const_scale_ext<J, const S: i32>(
self,
rhs: ConstScaleFpdec<J, S>,
rounding: Rounding,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
pub fn checked_mul_const_scale_ext<J, const S: i32>(
self,
rhs: ConstScaleFpdec<J, S>,
rounding: Rounding,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
Checked multiplication with ConstScaleFpdec. Computes self * rhs,
returning None if overflow occurred.
Compared to Self::checked_mul_ext, the result of this method inherits
the scale of self. So the diff_scale (which is equal to S) is no
more need.
Examples
use primitive_fixed_point_decimal::{ConstScaleFpdec, OobScaleFpdec, Rounding, fpdec};
type Balance = OobScaleFpdec<i64>;
type FeeRate = ConstScaleFpdec<i16, 4>;
let balance: Balance = fpdec!(12.30, 2); // scale=2
let rate: FeeRate = fpdec!(0.01);
let fee: Balance = balance.checked_mul_const_scale(rate).unwrap();
assert_eq!(fee, fpdec!(0.12, 2));
let fee: Balance = balance.checked_mul_const_scale_ext(rate, Rounding::Ceiling).unwrap();
assert_eq!(fee, fpdec!(0.13, 2));Sourcepub fn checked_div<J>(
self,
rhs: OobScaleFpdec<J>,
diff_scale: i32,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
pub fn checked_div<J>(
self,
rhs: OobScaleFpdec<J>,
diff_scale: i32,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
Checked division.
Equivalent to Self::checked_div_ext with Rounding::Round.
Sourcepub fn checked_div_ext<J>(
self,
rhs: OobScaleFpdec<J>,
diff_scale: i32,
rounding: Rounding,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
pub fn checked_div_ext<J>(
self,
rhs: OobScaleFpdec<J>,
diff_scale: i32,
rounding: Rounding,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
Checked division. Computes self / rhs, returning None if
division by 0, or overflow occurred, or argument diff_scale
is out of range [-Self::DIGITS, Self::DIGITS].
The type of rhs can have different inner integer J,
while the type of result must have the same I.
Argument: diff_scale = scale(self) - scale(rhs) - scale(result).
You can specify the rounding type.
Examples
use primitive_fixed_point_decimal::{OobScaleFpdec, Rounding, fpdec};
type Balance = OobScaleFpdec<i64>;
type FeeRate = OobScaleFpdec<i16>; // different types
let fee: Balance = fpdec!(0.13, 2); // scale=2
let rate: FeeRate = fpdec!(0.03, 4); // scale=4
let balance: Balance = fee.checked_div_ext(rate, -4, Rounding::Ceiling).unwrap();
assert_eq!(balance, fpdec!(4.34, 2));Sourcepub fn checked_div_const_scale<J, const S: i32>(
self,
rhs: ConstScaleFpdec<J, S>,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
pub fn checked_div_const_scale<J, const S: i32>(
self,
rhs: ConstScaleFpdec<J, S>,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
Checked division with ConstScaleFpdec.
Equivalent to Self::checked_div_const_scale_ext with Rounding::Round.
If you make sure no overflow or 0-division error, you can use / and /=
instead for convenience.
Sourcepub fn checked_div_const_scale_ext<J, const S: i32>(
self,
rhs: ConstScaleFpdec<J, S>,
rounding: Rounding,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
pub fn checked_div_const_scale_ext<J, const S: i32>(
self,
rhs: ConstScaleFpdec<J, S>,
rounding: Rounding,
) -> Option<OobScaleFpdec<I>>where
J: FpdecInner,
Checked division with ConstScaleFpdec. Computes self / rhs,
returning None if division by 0 or overflow occurred.
Compared to Self::checked_div_ext, the result of this method inherits
the scale of self. So the diff_scale (which is equal to -S) is no
more need.
Examples
use primitive_fixed_point_decimal::{ConstScaleFpdec, OobScaleFpdec, Rounding, fpdec};
type Balance = OobScaleFpdec<i64>;
type FeeRate = ConstScaleFpdec<i16, 4>;
let fee: Balance = fpdec!(0.12, 2); // scale=2
let rate: FeeRate = fpdec!(0.01);
let balance: Balance = fee.checked_div_const_scale(rate).unwrap();
assert_eq!(balance, fpdec!(12.0, 2));Sourcepub fn round_diff(self, diff_scale: i32) -> Self
pub fn round_diff(self, diff_scale: i32) -> Self
Round the decimal.
Equivalent to Self::round_diff_ext with Rounding::Round.
Sourcepub fn round_diff_ext(self, diff_scale: i32, rounding: Rounding) -> Self
pub fn round_diff_ext(self, diff_scale: i32, rounding: Rounding) -> Self
Round the decimal with rounding type.
The argument diff_scale is original_scale - round_scale.
Return the original decimal if diff_scale <= 0.
Examples:
use primitive_fixed_point_decimal::{OobScaleFpdec, Rounding, fpdec};
type Price = OobScaleFpdec<i64>;
let price: Price = fpdec!(12.12345678, 8); // scale=8
assert_eq!(price.round_diff(8 - 6), // reduce 2 scale
fpdec!(12.123457, 8)); // `Rounding::Round` as default
assert_eq!(price.round_diff_ext(8 - 6, Rounding::Floor),
fpdec!(12.123456, 8));Sourcepub fn try_from_str(s: &str, scale: i32) -> Result<Self, ParseError>where
I: Num<FromStrRadixErr = ParseIntError>,
pub fn try_from_str(s: &str, scale: i32) -> Result<Self, ParseError>where
I: Num<FromStrRadixErr = ParseIntError>,
Read decimal from string.
This method has 2 limitations:
- Support decimal format only but not scientific notation;
- Return
ParseError::Precisionif the string has more precision.
If you want to skip these limitations, you can parse the string to float number first and then convert the number to this decimal.
Examples:
use primitive_fixed_point_decimal::{OobScaleFpdec, ParseError, fpdec};
type Decimal = OobScaleFpdec<i16>;
assert_eq!(Decimal::try_from_str("1.23", 4).unwrap(), fpdec!(1.23, 4));
assert_eq!(Decimal::try_from_str("9999", 4), Err(ParseError::Overflow));
assert_eq!(Decimal::try_from_str("1.23456", 4), Err(ParseError::Precision));Sourcepub fn to_f32(self, scale: i32) -> f32
pub fn to_f32(self, scale: i32) -> f32
Convert into f32.
Examples:
use primitive_fixed_point_decimal::{OobScaleFpdec, fpdec};
type Decimal = OobScaleFpdec<i32>;
let dec: Decimal = fpdec!(1.234, 4);
assert_eq!(dec.to_f32(4), 1.234);
let dec: Decimal = fpdec!(1234000, -3);
assert_eq!(dec.to_f32(-3), 1234000.0);Sourcepub fn to_f64(self, scale: i32) -> f64
pub fn to_f64(self, scale: i32) -> f64
Convert into f64.
Examples:
use primitive_fixed_point_decimal::{OobScaleFpdec, fpdec};
type Decimal = OobScaleFpdec<i32>;
let dec: Decimal = fpdec!(1.234, 4);
assert_eq!(dec.to_f64(4), 1.234);
let dec: Decimal = fpdec!(1234000, -3);
assert_eq!(dec.to_f64(-3), 1234000.0);Sourceimpl<I> OobScaleFpdec<I>where
I: FpdecInner + Signed,
impl<I> OobScaleFpdec<I>where
I: FpdecInner + Signed,
Sourcepub fn abs(self) -> Self
pub fn abs(self) -> Self
Computes the absolute value of self.
Overflow behavior
The absolute value of MIN cannot be represented as this type,
and attempting to calculate it will cause an overflow. This means that
code in debug mode will trigger a panic on this case and optimized code
will return MIN without a panic.
Sourcepub fn checked_abs(self) -> Option<Self>
pub fn checked_abs(self) -> Option<Self>
Checked absolute value. Computes self.abs(), returning None if self == MIN.